"""Interface to the world 3 simulator."""
import dataclasses
import os

import numpy as np
from py_mini_racer import py_mini_racer

import whynot as wn
from whynot.dynamics import BaseConfig, BaseState, BaseIntervention


# Load javascript file for execution
DIR_NAME = os.path.dirname(__file__)
with open(os.path.join(DIR_NAME, "world3_app.js")) as handle:
    WORLD3_JS_CODE = handle.read()


# This is a hack to avoid a deadlock issue that
# arises when concurrently executing many MiniRacerContexts.
# There's some sort of issue with how the underlying v8 executor
# refers to contexts. Execution is thread-safe, but
# should fully sort this out before final release.
class PyMiniRacerContext(py_mini_racer.MiniRacer):
    # pylint: disable-msg=too-few-public-methods
    """Create an PyMiniRacer execution context."""

    def __del__(self):
        """Do nothing on deletion to avoid clobbering other processes."""


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """World3 simulation dynamics parameters.

    Default values correspond to the standard run of World3.

    Examples
    --------
    >>> # Configuration of a run from 1910-1950
    >>> world3.Config(start_time=1910, end_time=1950)

    .. note:
        There are many possible parameters to adjust in World3. This subset
        exposed in the Config correspond to variables that are scalar quantities
        rather than tabular functions for simplicity.

    """

    # Dynamics parameters. The #: ensures the attribute can be documented by Sphinx.
    #:
    industrial_capital_output_ratio: float = 3
    #:
    average_lifetime_of_industrial_capital: float = 14
    #:
    fraction_of_industrial_output_allocated_to_consumption_constant: float = 0.43
    #:
    average_lifetime_of_service_capital: float = 20
    #:
    service_capital_output_ratio: float = 1
    #:
    land_yield_factor: float = 1
    #:
    nonrenewable_resource_usage_factor: float = 1
    #:
    persistent_pollution_generation_factor: float = 1

    # Simulator parameters
    #: Year to start the simulation.
    start_time: float = 1900
    #: Year to end the simulation.
    end_time: float = 2100
    #: Granularity of the simulation (step size in the forward Euler method).
    delta_t: float = 1.0


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the World3 model.

    An intervention is a subset of the configuration variables,
    and only variables passed into the constructor are modified.

    Examples
    --------
    >>> # Semantics: Starting in year 1975, double land_yield. All
    >>> # other parameters are unchanged.
    >>> Intervention(time=1975, land_yield_factor=2)

    """

    def __init__(self, time=1975, **kwargs):
        """Construct an intervention object.

        Parameters
        ----------
            time: float
                time (in years) to intervene in the simulator.
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """World3 state.

    Default values correspond to the initial state in 1900 of the standard run.

    Examples
    --------
    >>> # Construct initial state with 800 land fertility.
    >>> world3.State(arable_land=800)

    """

    #:
    population_0_to_14: float = 6.5e8
    #:
    population_15_to_44: float = 7.0e8
    #:
    population_45_to_64: float = 1.9e8
    #:
    population_65_and_over: float = 6.0e7
    #:
    industrial_capital: float = 2.1e11
    #:
    service_capital: float = 1.44e11
    #:
    arable_land: float = 0.9e9
    #:
    potentially_arable_land: float = 2.3e9
    #:
    urban_industrial_land: float = 8.2e6
    #:
    land_fertility: float = 600.0
    #:
    nonrenewable_resources: float = 1.0e12
    #:
    persistent_pollution: float = 2.5e7

    @property
    def total_population(self):
        """Return the aggregate population."""
        return (
            self.population_0_to_14
            + self.population_15_to_44
            + self.population_45_to_64
            + self.population_65_and_over
        )


def to_camel_case(snake_str):
    """Convert snake_str to snakeStr (camel case)."""
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])


def set_state(js_context, initial_state):
    """Set the state of the world3 simulator."""
    for stock_name, value in dataclasses.asdict(initial_state).items():
        js_context.eval(f"{to_camel_case(stock_name)}.initVal = {value}")

    # special case for resources
    js_context.eval(
        f"nonrenewableResourcesInitialK = {initial_state.nonrenewable_resources}"
    )
    js_context.eval("resetModel()")


def decode_states(js_context):
    """Read out the sequence of states from the world3 engine.

    Parameters
    ----------
        js_context: PyMiniRacerContext
            Context containing a completed execution of world3.

    Returns
    -------
        (states, sample_times): list, np.ndarray
            The recorded state values and the time each value was sampled during the run.

    """

    def unpack(values):
        """Parse list of [{"x": time, "y":value}] into [times], [values]."""
        return list(zip(*[(value["x"], value["y"]) for value in values]))

    state_names = [f.name for f in dataclasses.fields(State)]
    data = {}
    sample_times = None
    for idx, state_name in enumerate(state_names):
        state_data = js_context.eval(f"{to_camel_case(state_name)}.data")
        times, values = unpack(state_data)
        if idx == 0:
            sample_times = times
        data[state_name] = values

    states = [State(*vals) for vals in zip(*data.values())]

    return states, np.array(sample_times)


def set_config(js_context, config, intervention):
    """Set the non-state variables of the world3 simulator."""
    # Set global simulator parameters
    js_context.eval(f"startTime = {config.start_time}")
    js_context.eval(f"stopTime = {config.end_time}")
    js_context.eval(f"dt = {config.delta_t}")

    if intervention:
        intervention_config = config.update(intervention)
        js_context.eval(f"policyYear = {intervention.time}")
    else:
        intervention_config = config
        js_context.eval(f"policyYear = {config.end_time}")

    intervention_config = dataclasses.asdict(intervention_config)
    for parameter, before in dataclasses.asdict(config).items():
        if parameter in ["policy_year", "start_time", "end_time", "delta_t"]:
            continue
        after = intervention_config[parameter]
        js_context.eval(f"{to_camel_case(parameter)}.before = {before}")
        js_context.eval(f"{to_camel_case(parameter)}.after = {after}")
    js_context.eval("resetModel()")


def simulate(initial_state, config, intervention=None, seed=None):
    """Run the world3 simulation for the specified initial state and configuration.

    Uses the forward Euler method internally to simulate world3 from start_time
    to end_time (specified in config) with a step size of delta_t (in Config).

    Parameters
    ----------
        initial_state: whynot.simulators.world3.State
            Initial state of the dynamical system
        config: whynot.simulators.world3.Config
            Configuraton parameters to control simulator dynamics.
        intervention: whynot.simulators.world3.Intervention
            (Optional) Intervention, if any, to perform during simulator execution.
        seed: int
            (Optional) Ignored since the simulator is deterministic.

    Returns
    -------
        run: whynot.dynamics.Run
            A single rollout from the simulator.

    """
    # pylint:disable-msg=unused-argument
    # Initialize a separate context separately for each simulator rollout.
    # This may be potentially less efficient that a single global context,
    # but it ensures that parallel executions will work without issue.
    ctx = PyMiniRacerContext()

    # Load the world3 simulation code
    ctx.eval(WORLD3_JS_CODE)

    # Initialize the similuator
    set_state(ctx, initial_state)
    set_config(ctx, config, intervention)

    # Single simulator rollout
    ctx.eval("fastRun()")

    # Read out the results
    states, times = decode_states(ctx)
    # Ensure the user passed initial state appears in the run
    states = [initial_state] + states[1:]
    return wn.dynamics.Run(states=states, times=times)



"""Experiments on the DICE model."""
import numpy as np

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import dice

__all__ = ["get_experiments", "RCT"]


def get_experiments():
    """Return all of the Lotka-Volterra experiments."""
    return [RCT]


def sample_initial_states():
    """Sample an initial state for the DICE model."""
    # Initial state is currently deterministic. Randomness
    # comes from stochastic dynamics.
    return dice.State()


def dice_outcome_extractor(run):
    """Return the atmospheric temperature of the final state."""
    return run.states[-1].TATM


###########################
# RCT Experiment
###########################
@parameter(
    name="propensity",
    default=0.5,
    values=np.linspace(0.05, 0.5, 10),
    description="Probability of treatment",
)
def rct_propensity(propensity):
    """Return constant propensity for RCT."""
    return propensity


#: An RCT probing the effect of using optimal carbon prices on atmospheric temperature
RCT = DynamicsExperiment(
    name="dice_rct",
    description="A RCT to determine effect of optimal carbon prices on atmospheric temperature.",
    simulator=dice,
    simulator_config=dice.Config(ifopt=0, numPeriods=10),
    intervention=dice.Intervention(ifopt=1),
    state_sampler=sample_initial_states,
    propensity_scorer=rct_propensity,
    outcome_extractor=dice_outcome_extractor,
    covariate_builder=lambda run: run.initial_state.values(),
)



"""Utility functions used by all of the simulators."""
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
import copy
import itertools

import numpy as np
from tqdm.auto import tqdm

import whynot as wn


def pretty_print(experiment, dataset, results):
    """Print the results of running the causal suite on data from an experiment.

    Parameters
    ----------
        experiment: `whynot.dynamics.DynamicsExperiment` or `whynot.framework.GenericExperiment`
            Experiment object used to generate the causal dataset
        dataset: `whynot.framework.Dataset`
            Dataset object passed to the causal suite.
        results:    dict
            Dictionary of results returned running `whynot.causal_suite` on the dataset

    """
    print("Name: ", experiment.name)
    print("Description: ", experiment.description)
    for method, estimate in results.items():
        print(f"Method: {method:<25} \t\t Estimate: {estimate.ate:2.2e}")
    print(f"{' ':<30} \t\t\t Ground Truth: {dataset.sate:2.2e}")


def parallelize(func, arg_lst, show_progress=False, max_workers=None):
    """Parallel execution of function func across a list of arguments.

    The function func and all of the arguments must be pickable. Func is
    executed on each elements of arg_list as func(*args)

    Parameters
    ----------
        func:
            Function to repeatedly execute, must be pickable.
        arg_lst: iterable
            Iterator of unnamed arguments. Each element arg is passed as func(*arg).
        show_progress: bool
            Whether or not to display a progress bar.
        max_workers: int
            Maximum number of parallel processes to execute simultaneously.

    Returns
    -------
        results: list
            List of outcomes of running func(*arg) on each arg in arg_list.
            Results are in the same order as the input arg_list.

    """

    def display(range_obj):
        if show_progress:
            range_obj = tqdm(range_obj)
        return range_obj

    results = []
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for args in arg_lst:
            futures.append(executor.submit(func, *args))
        for future in display(futures):
            data = future.result()
            results.append(data)
    return results


def parallel_run_estimators(causal_datasets):
    """Run causal suite in parallel for repeated trials of a causal experiment.

    Parameters
    ----------
        causal_datasets: dict
            Dictionary mapping an experiment setting to a list of datasets
            representing repeated trials of the experiment.

    Returns
    -------
        all_estimates: dict
            Dictionary mapping estimators name and experimental setting
            to a list of `wn.InferenceResult` objects for each trial,
            e.g. all_estimates['ols'][200][3] is the InferenceResult for ols
            on the 3rd trial of the experiment with setting 200.

    """
    all_estimates = defaultdict(lambda: defaultdict(list))
    for key, trials in causal_datasets.items():
        parallel_args = [
            (dataset.covariates, dataset.treatments, dataset.outcomes)
            for dataset in trials
        ]
        all_trial_estimates = parallelize(
            wn.causal_suite, parallel_args, show_progress=True
        )
        for estimates in all_trial_estimates:
            for method, estimate in estimates.items():
                all_estimates[method][key].append(estimate)
    return all_estimates


def sample_size_experiment(
    experiment, sample_sizes, num_trials, parameters=None, seeds=None, verbose=False
):
    """Repeatedly run an experiment at different sample sizes.

    All of the datasets are generate sequentially, and the estimators are run in
    parallel.

    Parameters
    ----------
        experiment: `whynot.dynamics.DynamicsExperiment` or `whynot.framework.GenericExperiment`
            Instantiated experiment object.
        sample_sizes: list
            List of sample sizes to run the experiment
        num_trials: int
            How many trials to run each experiment for a fixed sample size.
        parameters: dict
            Dictionary of {param_name: param_value} fixing non-varying parameters
            for the experiment.
        (optional) seeds: list
           List of random seeds to use for each trial. If specified, should have
           length num_trials.
        (optional) verbose: bool
            Print status updates.

    Returns
    -------
        estimates: dict
            Dictionary mapping each method to a dictionary of sample_size to
            `whynot.framework.InferenceResults` for each trial at the given
            sample size.

                estimates[method_name] = {
                    sample_size1: [estimates_for_sample_size_1],
                    sample_size2: [estimates_for_sample_size_2],
                    ...}

        sample_ates: dict
            Dictionary mapping sample_size to the sample_ate for each trial.

                sample_ates = {
                    sample_size1: [sample_ates_for_sample_size_1],
                    sample_size2: [sample_ates_for_sample_size_2],
                    ...}

    """
    if seeds is None:
        seeds = [None] * num_trials
    assert len(seeds) == num_trials
    if parameters is None:
        parameters = {}

    if verbose:
        print("Generating causal datasets...")
    datasets = defaultdict(list)
    all_sates = defaultdict(list)
    for (sample_size, seed) in itertools.product(sample_sizes, seeds):
        dataset = experiment.run(num_samples=sample_size, seed=seed, **parameters)
        all_sates[sample_size].append(dataset.sate)
        datasets[sample_size].append(dataset)

    if verbose:
        print("Running estimators...")
    return parallel_run_estimators(datasets), all_sates


def parameter_sweep_experiment(
    experiment,
    sample_size,
    num_trials,
    parameter_name,
    parameter_values,
    fixed_parameters=None,
    seeds=None,
    verbose=False,
):
    """Repeatedly run an experiment for different values of a parameter.

    All of the datasets are generate sequentially, and the estimators are run in
    parallel.

    Parameters
    ----------
        experiment: `whynot.dynamics.DynamicsExperiment` or `whynot.framework.GenericExperiment`
            Instantiated experiment object.
        sample_size: int
            Sample size to use for all experiments.
        num_trials: int
            How many trials to run each experiment for a fixed parameter setting.
        parameter_name: str
            Name of the parameter to vary.
        parameter_values: list
            List of values of the parameter to vary
        fixed_parameters: dict
            Dictionary of {param_name: param_value} fixing non-varying parameters
            for the experiment.
        (optional) seeds: list
           List of random seeds to use for each trial. If specified, should have
           length num_trials.
        (optional) verbose: bool
            Print status updates.

    Returns
    -------
        estimates: dict
            Dictionary mapping each method to a dictionary of parameter_value to
            `whynot.framework.InferenceResults` for each trial at the given
            sample size.

                estimates[method_name] = {
                    parameter_value1: [estimates_for_parameter_value1],
                    parameter_value2: [estimates_for_parameter_value2],
                    ...}

        sample_ates: dict
            Dictionary mapping parameter_value to the sample_ate for each trial.

    """
    if seeds is None:
        seeds = [None] * num_trials
    assert len(seeds) == num_trials
    if fixed_parameters is None:
        fixed_parameters = {}

    if verbose:
        print("Generating causal datasets...")
    datasets = defaultdict(list)
    sample_ates = defaultdict(list)
    for (parameter_value, seed) in itertools.product(parameter_values, seeds):
        parameters = copy.deepcopy(fixed_parameters)
        parameters[parameter_name] = parameter_value
        dataset = experiment.run(num_samples=sample_size, seed=seed, **parameters)
        sample_ates[parameter_value].append(dataset.sate)
        datasets[parameter_value].append(dataset)
    if verbose:
        print("Running estimators...")
    return parallel_run_estimators(datasets), sample_ates


def summarize_errors(estimates, sample_ates, metric):
    """Summarize estimator errors for a parameter or sample size sweep.

    Currently, this function only supports summaries for ATE estimation.
    This function should be used in conjunction with parameter_sweep_experiment
    and sample_size_experiment.

    Parameters
    ----------
        estimates: dict
            Dictionary mapping method_names to a dictionary of experiment
            settings and `whynot.InferenceResults` as returned by
            parameter_sweep_experiment.
        sample_ates: dict
            Dictionary mapping experiment settings to sample ates.
        metric: str
            One of 'relative_error' or 'absolute_error' for reporting results.

    Returns
    -------
        summary: dict
            Dictionary mapping method name to a tuple of (means, stds), where
            means is a list of mean error for each experimental setting, and
            similarly for standard deviation.

    """

    def score(est, sate):
        if metric == "relative_error":
            return np.abs((est - sate) / sate)

        if metric == "absolute_error":
            return np.abs(est - sate)

        raise NotImplementedError

    summary = {}
    for method, results in estimates.items():
        means, stds = [], []
        for setting, inferences in results.items():
            scores = []
            for inference, sample_ate in zip(inferences, sample_ates[setting]):
                scores.append(score(inference.ate, sample_ate))
            means.append(np.mean(scores))
            stds.append(np.std(scores) / np.sqrt(len(scores)))
        summary[method] = (means, stds)
    return summary



"""Global registry of environments, for consistency with openai gym."""
import importlib

from gym.envs.registration import EnvRegistry

# Keep for consistency with original API
# pylint:disable-msg=invalid-name
# Have a global registry
registry = EnvRegistry()


# pylint:disable-msg=redefined-builtin
def register(id, **kwargs):
    """Register the environment."""
    return registry.register(id, **kwargs)


def make(id, **kwargs):
    """Build the environment."""
    return registry.make(id, **kwargs)


def spec(id):
    """View the spec for the environment."""
    return registry.spec(id)


warn_once = True



"""Implementation of the Perdomo et. al model of strategic classification.

The data is from the Kaggle Give Me Some Credit dataset:

    https://www.kaggle.com/c/GiveMeSomeCredit/data,

and the dynamics are taken from:

    Perdomo, Juan C., Tijana Zrnic, Celestine Mendler-Dünner, and Moritz Hardt.
    "Performative Prediction." arXiv preprint arXiv:2002.06673 (2020).
"""
import copy
import dataclasses
from typing import Callable, List

import whynot as wn
import whynot.traceable_numpy as np
from whynot.dynamics import BaseConfig, BaseIntervention, BaseState
from whynot.simulators.credit.dataloader import CreditData


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of Credit simulator dynamics.

    Examples
    --------
    >>> # Configure simulator for run for 10 iterations
    >>> config = Config(start_time=0, end_time=10, delta_t=1)

    """

    # Dynamics parameters
    #: Subset of the features that can be manipulated by the agent
    changeable_features: np.ndarray = np.array([0, 5, 7])

    #: Model how much the agent adapt her features in response to a classifier
    epsilon: float = 0.1

    #: Parameters for logistic regression classifier used by the institution
    theta: np.ndarray = np.ones((11, 1))

    #: L2 penalty on the logistic regression loss
    l2_penalty: float = 0.0

    #: Whether or not dynamics have memory
    memory: bool = False

    # Simulator book-keeping
    #: Start time of the simulator
    start_time: int = 0
    #: End time of the simulator
    end_time: int = 5
    #: Spacing of the evaluation grid
    delta_t: int = 1


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the Credit model."""

    #: Matrix of agent features (see https://www.kaggle.com/c/GiveMeSomeCredit/data)
    features: np.ndarray = np.ones((13, 11))

    #: Vector indicating whether or not the agent experiences financial distress
    labels: np.ndarray = np.zeros((13, 1))

    def values(self):
        """Return the state as a dictionary of numpy arrays."""
        return {name: getattr(self, name) for name in self.variable_names()}


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the Credit model.

    An intervention changes a subset of the configuration variables in the
    specified year. The remaining variables are unchanged.

    Examples
    --------
    >>> # Starting at time 25, update the classifier to random chance.
    >>> config = Config()
    >>> Intervention(time=25, theta=np.zeros_like(config.theta))

    """

    def __init__(self, time=30, **kwargs):
        """Specify an intervention in credit.

        Parameters
        ----------
            time: int
                Time of intervention in simulator dynamics.
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


def strategic_logistic_loss(config, features, labels, theta):
    """Evaluate the performative loss for logistic regression classifier."""

    config = config.update(Intervention(theta=theta))

    # Compute adjusted data
    strategic_features = agent_model(features, config)

    # compute log likelihood
    num_samples = strategic_features.shape[0]
    logits = strategic_features @ config.theta
    log_likelihood = (1.0 / num_samples) * np.sum(
        -1.0 * np.multiply(labels, logits) + np.log(1 + np.exp(logits))
    )

    # Add regularization (without considering the bias)
    regularization = (config.l2_penalty / 2.0) * np.linalg.norm(config.theta[:-1]) ** 2

    return log_likelihood + regularization


def agent_model(features, config):
    """Compute agent reponse to the classifier and adapt features accordingly.

    TODO: For now, the best-response model corresponds to best-response with
    linear utility and quadratic costs. We should expand this to cover a rich
    set of agent models beyond linear/quadratic, and potentially beyond
    best-response.
    """
    # Move everything by epsilon in the direction towards better classification
    strategic_features = np.copy(features)
    theta_strat = config.theta[config.changeable_features].flatten()
    strategic_features[:, config.changeable_features] -= config.epsilon * theta_strat
    return strategic_features


def dynamics(state, time, config, intervention=None):
    """Perform one round of interaction between the agents and the credit scorer.

    Parameters
    ----------
        state: whynot.simulators.credit.State
            Agent state at time TIME
        time: int
            Current round of interaction
        config: whynot.simulators.credit.Config
            Configuration object controlling the interaction, e.g. classifier
            and agent model
        intervention: whynot.simulators.credit.Intervention
            Intervention object specifying when and how to update the dynamics.

    Returns
    -------
        state: whynot.simulators.credit.State
            Agent state after one step of strategic interaction.

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    if config.memory:
        features, labels = state
    else:
        features, labels = CreditData.features, CreditData.labels

    # Update features in response to classifier. Labels are fixed.
    strategic_features = agent_model(features, config)
    return strategic_features, labels


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Credit model.

    Parameters
    ----------
        initial_state: whynot.credit.State
        config: whynot.credit.Config
            Base parameters for the simulator run
        intervention: whynot.credit.Intervention
            (Optional) Parameters specifying a change in dynamics
        seed: int
            Unused since the simulator is deterministic.

    Returns
    -------
        run: whynot.dynamics.Run
            Simulator rollout

    """
    # Iterate the discrete dynamics
    times = [config.start_time]
    states = [initial_state]
    state = copy.deepcopy(initial_state)

    for step in range(config.start_time, config.end_time):
        next_state = dynamics(state.values(), step, config, intervention)
        state = State(*next_state)
        states.append(state)
        times.append(step + 1)

    return wn.dynamics.Run(states=states, times=times)


if __name__ == "__main__":
    print(simulate(State(), Config(end_time=2)))



"""Load and preprocess Kaggle credit dataset."""
import os

import numpy as np
import pandas as pd
from sklearn import preprocessing


class DataLoader(object):
    """Class to lazily load the credit dataset."""

    def __init__(self, datafile="credit_data.zip", seed=None):
        cur_dir = os.path.abspath(os.path.dirname(__file__))
        datapath = os.path.join(cur_dir, datafile)

        self.datapath = datapath
        self._features = None
        self._labels = None
        self.seed = seed

    @property
    def features(self):
        """Return the dataset features."""
        if self._features is None:
            self._features, self._labels = self.load_data()
        return np.copy(self._features)

    @property
    def labels(self):
        """Return the dataset labels."""
        if self._labels is None:
            self._features, self._labels = self.load_data()
        return np.copy(self._labels)

    @property
    def num_agents(self):
        """Compute number of agents in the dataset."""
        return self.features.shape[0]

    @property
    def num_features(self):
        """Compute number of features for each agent."""
        return self.features.shape[1]

    def load_data(self):
        """Load, preprocess and class-balance the credit data."""
        rng = np.random.RandomState(self.seed)

        data = pd.read_csv(self.datapath, index_col=0)
        data.dropna(inplace=True)

        features = data.drop("SeriousDlqin2yrs", axis=1)
        # zero mean, unit variance
        features = preprocessing.scale(features)

        # add bias term
        features = np.append(features, np.ones((features.shape[0], 1)), axis=1)
        outcomes = np.array(data["SeriousDlqin2yrs"])

        # balance classes
        default_indices = np.where(outcomes == 1)[0]
        other_indices = np.where(outcomes == 0)[0][:10000]
        indices = np.concatenate((default_indices, other_indices))

        features_balanced = features[indices]
        outcomes_balanced = outcomes[indices]

        shape = features_balanced.shape

        # shuffle arrays
        shuffled = rng.permutation(len(indices))
        return features_balanced[shuffled], outcomes_balanced[shuffled]


CreditData = DataLoader(seed=0)



"""Base experimental class."""
import dataclasses
import inspect
import itertools
from typing import Any

import numpy as np


PARAM_COLLECTION = "__parameter_collection__"


@dataclasses.dataclass
class Dataset:
    # pylint:disable-msg=too-few-public-methods
    """Observational dataset and grouth truth unit-level effects.

    Attributes
    ----------
        covariates: `np.ndarray`
            Float array of shape [num_samples, num_features] of covariates for each unit.
        treatments:  `np.ndarray`
            Integer 0/1 array of shape [num_samples] indicating treatment
            status for each unit. 1 indicates treated, 0 indicates unit not treated.
        outcomes:  `np.ndarray`
            Float array of shape [num_samples] containing the observed outcome for each unit.
        true_effects: `np.ndarray`
            Float array of shape [num_samples] containing the unit-level treatment
            effects, :math:`Y_i(1) - Y_i(0)` for each :math:`i`.
        sate: float
            Sample average treatment effect based on ground truth unit-level effects.
        causal_graph: networkx.DiGraph
            If supported by the simulator and experiment, the causal graph associated with the data.

    """

    covariates: np.ndarray
    treatments: np.ndarray
    outcomes: np.ndarray
    true_effects: np.ndarray
    causal_graph: Any = None

    @property
    def sate(self):
        """Return the sample average treatment effect."""
        return np.mean(self.true_effects)

    def bootstrap_sample_ate_ci(self, num_bootstrap_samples=2000, alpha=0.05):
        """Bootstrap a (1-alpha)% confidence interval for the sample ate."""
        means = []
        for _ in range(num_bootstrap_samples):
            sample = np.random.choice(
                self.true_effects, size=len(self.true_effects), replace=True
            )
            means.append(np.mean(sample))
        lower_tail, upper_tail = alpha / 2.0, 1.0 - alpha / 2.0
        return (np.quantile(means, lower_tail), np.quantile(means, upper_tail))


@dataclasses.dataclass
class InferenceResult:
    # pylint:disable-msg=too-few-public-methods
    """Object to store results of causal inference method.

    Attributes
    ----------
        ate:
            Estimated average treatment effect
        stderr:
            Reported standard error of the ATE estimate. Only available
            if supported by the method, otherwise None.
        ci:
            Reported 95% confidence interval for the ATE (lower_bound, upper_bound).
            Only available if supported by the method, otherwise None.
            TODO: Ideally, we'd support various significance levels.
        individual_effects:
            Heterogeneous treatment effect for each unit.
            Only available if supported by the method, otherwise None.
        elapsed_time:
            How long in second (wall-clock time) it took to produce the estimate.

    """

    ate: float = None
    stderr: float = None
    ci: tuple = (None, None)  # pylint: disable-msg=invalid-name
    individual_effects: np.ndarray = None
    elapsed_time: float = None


@dataclasses.dataclass
class ExperimentParameter:
    # pylint:disable-msg=too-few-public-methods
    """Container for a parameter to vary in an experiment.

    Attributes
    ----------
        name:
            Name of the parameter
        description:
            Parameter description
        default:
            Default (uninitialized) value of the parameter
        values:
            Iterator of parameter values that supports sampling (for random search).

    """

    name: str
    default: Any
    values: Any = None
    description: str = ""


def parameter(name, default, values=None, description=""):
    """Decorate functions in an experiment with parameters.

    Usage:
        @parameter(name="treatment-bias", default=0.2, values=[0.1, 0.2, 0.5, 0.9],
                   description="How much is treatment biased for control group.")
        def propensity_score(untreated_run, treatment_bias)
            ....
    """
    exp_param = ExperimentParameter(name, default, values, description)

    def parameter_decorator(func):
        """Attach parameter collection object to the function."""
        # Ensure the parameter is an argument to the function
        method_params = inspect.signature(func).parameters
        if name not in method_params:
            raise ValueError(f"{name} is not an argument to {func.__name__}")

        if hasattr(func, PARAM_COLLECTION):
            getattr(func, PARAM_COLLECTION).add_parameter(exp_param)
        else:
            setattr(func, PARAM_COLLECTION, ParameterCollection([exp_param]))

        return func

    return parameter_decorator


def extract_params(func, standard_args):
    """Return WhyNot parameters for user-defined function.

    Performs error-checking to ensure parameters are disjoint from
    standard arguments and all arguments to the function are either
    standard arguments or parameters.

    Parameters
    ----------
        func: function
            Possibly parameterized function.

        standard_args: list
            A list of possible arguments provided by the calling class itself
            and shouldn't be treated as parameters for the function.

    Returns
    -------
        params: `whynot.framework.ParameterCollection`
            A collection of parameters for the func.

    """
    if not callable(func):
        msg = f"Trying to extract parameters from {func.__name__}, but not callable."
        raise ValueError(msg)

    # Extract parameters specified by the user via the @parameter decorator.
    specified_params = ParameterCollection([])
    if hasattr(func, PARAM_COLLECTION):
        specified_params = getattr(func, PARAM_COLLECTION)

    # Ensure standard_args is disjoint from the specified params.
    for arg in standard_args:
        if arg in specified_params:
            msg = (
                f"{arg} is both a parameter and a standard argument to {func.__name__}."
            )
            raise ValueError(msg)

    # By construction, every element in specified_params
    # must appear in the function signature, i.e. method_params.
    method_args = inspect.signature(func).parameters
    for arg in method_args:
        if arg not in standard_args and arg not in specified_params:
            msg = (
                f"'{arg}' is in the signature of function {func.__name__}, "
                f"but '{arg}' is not a standard argument or a parameter. "
                f"Standard arguments: {', '.join(standard_args)}."
            )
            raise ValueError(msg)

    return specified_params


class ParameterCollection:
    """Lightweight wrapper class around a set of parameters.

    Provides utility functions to support sampling and assigning subsets of the
    parameters.

    Enforces name uniqueness. Every parameter should have a unique name.

    """

    def __init__(self, params=None):
        """Params is a list of Parameter objects."""
        self.params = {}
        if params is not None:
            for param in params:
                if param.name in self.params:
                    raise ValueError(f"Duplicate name {param.name}")
                self.params[param.name] = param

    def add_parameter(self, param):
        """Add a new parameter to the existing collection."""
        if param.name in self.params:
            raise ValueError(f"Adding parameter with duplicate name {param.name}")
        self.params[param.name] = param

    def default(self):
        """Return the default parameter setting for each parameter."""
        return {name: p.default for name, p in self.params.items()}

    def sample(self, seed=None):
        """Return a random parameter setting for each parameter."""
        rng = np.random.RandomState(seed)
        sampled_params = {}
        for name, param in self.params.items():
            if param.values is None:
                sampled_params[name] = param.default
            else:
                sampled_params[name] = rng.choice(param.values)
        return sampled_params

    def project(self, specified_params):
        """Return fully instantiated parameters using defaults for unspecified params.

        specified_params is a dict of the arguments passed to the run method.
        Any unspecified params are set to their default values.

        Returns a dictonary params[param_name] -> param_value
        """
        for param_name in specified_params:
            if param_name not in self.params:
                raise ValueError(f"Parameter {param_name} specified, but not used!.")

        params = {}
        for name, param in self.params.items():
            if name in specified_params:
                params[name] = specified_params[name]
            else:
                params[name] = param.default

        return params

    def __contains__(self, name):
        """Check if param name is specified."""
        return name in self.params

    def __getitem__(self, name):
        """Return the parameter object corresponding to name."""
        if name in self.params:
            return self.params[name]
        raise ValueError(f"{name} not found")

    def __add__(self, collection):
        """Add two parameter collections together."""
        original_params = list(self.params.values())
        new_params = list(collection.params.values())
        return ParameterCollection(original_params + new_params)

    def __iter__(self):
        """Iterate over the collection."""
        for param in self.params.values():
            yield param

    def __repr__(self):
        """See ___str___."""
        return self.__str__()

    def __str__(self):
        """Display the collection in a human readable format.

        For instance:
        Params:
            Name:		    hidden_dim
            Description:	hidden dimension of 2-layer ReLu network response.
            Default:	    32
            Values:		    [8, 16, 32, 64, 128, 256, 512]
        """
        class_display = "Params:"
        param_display = (
            "\tName:\t\t{}\n\tDescription:\t{}\n\tDefault:\t{}\n\tValues:\t\t{}\n"
        )
        for param in self.params.values():
            param_values = [] if param.values is None else param.values
            class_display += "\n" + param_display.format(
                param.name, param.description, param.default, param_values
            )
        return class_display

    def grid(self):
        """Return a parameter grid.
        
        Examples
        --------
        >>> p1 = ExperimentParams(name="a", values=[1, 2])
        >>> p2 = ExperimentParams(name="b", values=[3, 4])
        >>> collection = ParameterCollection(params=[a, b])
        >>> for a, b in collection.grid():
        ...     print(a, b)
        1, 3
        2, 3
        1, 4
        2, 4
 
        """
        grid = []
        for values in itertools.product(*[p.values for p in self.params.values()]):
            settings = {}
            for param_, value in zip(self.params, values):
                settings[param_.name] = value
            grid.append(settings)
        return grid


####################
# Generic Simulator
####################
class GenericExperiment:
    # pylint:disable-msg=too-few-public-methods
    """Encapsulate a causal simulation experiment."""

    def __init__(self, name, description, run_method):
        """Initialize a generic experiment class."""
        self.name = name
        self.description = description
        self.run_method = run_method

        run_args = ["num_samples", "seed", "parallelize", "show_progress"]
        self.params = extract_params(run_method, run_args)

    def get_parameters(self):
        """Return parameters of the experiment."""
        return self.params

    def run(
        self, num_samples, seed, parallelize=True, show_progress=False, **parameter_args
    ):
        """Run the experiment and return a causal dataset."""
        run_parameters = self.params.project(parameter_args)
        results = self.run_method(
            num_samples=num_samples,
            seed=seed,
            show_progress=show_progress,
            parallelize=parallelize,
            **run_parameters,
        )

        (covariates, treatment, outcome), ground_truth = results
        return Dataset(
            covariates=covariates,
            treatments=treatment,
            outcomes=outcome,
            true_effects=ground_truth,
        )



"""Reinforcment learning for world3."""
from itertools import product

import numpy as np

from whynot.gym import spaces
from whynot.gym.envs import ODEEnvBuilder, register
from whynot.simulators.world3 import Config, Intervention, simulate, State


def get_intervention(action, time):
    """Return the intervention needed to take action in the simulator."""
    resource_usages = [0.8, 1.0, 1.2]
    pollution_generations = [0.8, 1.0, 1.2]
    action_list = list(product(resource_usages, pollution_generations))
    resource_use, pollution_gen = action_list[action]
    return Intervention(
        time,
        nonrenewable_resource_usage_factor=resource_use,
        persistent_pollution_generation_factor=pollution_gen,
    )


def get_reward(intervention, state):
    """Return the intervention needed to take action in the simulator."""
    # Constants are made up, but chosen so that all quantities have
    # the same magnitude for the initial state.

    # High population and capital are good, high pollution is bad.
    reward = 1e-11 * state.industrial_capital - 5e-7 * state.persistent_pollution
    reward += 2e-8 * state.total_population

    # Taking any action other than the "default" is costly
    resource_usage = intervention.updates["nonrenewable_resource_usage_factor"]
    pollution_generation = intervention.updates[
        "persistent_pollution_generation_factor"
    ]
    reward -= 0.3 * (1.0 - resource_usage) ** 2
    reward -= 0.5 * (1.0 - pollution_generation) ** 2

    return reward


def observation_space():
    """Return the model observation space."""
    num_states = State.num_variables()
    state_space_low = np.zeros(num_states)
    state_space_high = np.inf * np.ones(num_states)
    return spaces.Box(state_space_low, state_space_high)


World3Env = ODEEnvBuilder(
    simulate_fn=simulate,
    # Smaller delta_t improves numerical stability
    config=Config(delta_t=0.5),
    initial_state=State(),
    # In this environment there are 9 actions defined by
    # nonrenewable_resource_usage and pollution_generation_factor.
    action_space=spaces.Discrete(9),
    observation_space=observation_space(),
    timestep=1.0,
    intervention_fn=get_intervention,
    reward_fn=get_reward,
)

register(
    id="world3-v0", entry_point=World3Env, max_episode_steps=400, reward_threshold=1e5,
)
